@@ -2,11 +2,11 @@
 #
 # This file creates a w32 binary release
 #
-VERSION=1.01
+VERSION=1.21
 DDIR="`pwd`/Enigma-$VERSION"
 SDIR=..
 BDIR=..
-DLLDEFAULTS="SDL SDL_image SDL_mixer SDL_ttf jpeg zlib1 libfreetype-6 libpng12-0 mikmod"
+DLLDEFAULTS="SDL SDL_image SDL_mixer SDL_ttf jpeg zlib1 libfreetype-6 libogg-0 libpng12-0 libvorbis-0 libvorbisfile-3 mikmod libcurl-4"
 
 if [ "$ENIGMADEVDLLS" != "" ]; then
   DLLS=$ENIGMADEVDLLS
@@ -62,13 +62,20 @@ function copy_levels
     cplffiles data/levels data/levels lua
     cplffiles data/levels data/levels txt
     cplffiles data/levels data/levels xml
-    for folder in `(cd $SDIR/data/levels && ls -F)| grep / | grep -v CVS`; do 
+    for folder in `(cd $SDIR/data/levels && ls -F)| grep / `; do 
 	mkdir $DDIR/data/levels/$folder; 
 	cp -p $SDIR/data/levels/$folder/*.png $DDIR/data/levels/$folder
         cplffiles data/levels/$folder data/levels/$folder lua
         cplffiles data/levels/$folder data/levels/$folder txt
         cplffiles data/levels/$folder data/levels/$folder xml
     done
+    for folder in `(cd $SDIR/data/levels/soko && ls -F)| grep / `; do 
+	mkdir $DDIR/data/levels/soko/$folder; 
+	cp -p $SDIR/data/levels/soko/$folder/*.png $DDIR/data/levels/soko/$folder
+        cplffiles data/levels/soko/$folder data/levels/soko/$folder lua
+        cplffiles data/levels/soko/$folder data/levels/soko/$folder txt
+        cplffiles data/levels/soko/$folder data/levels/soko/$folder xml
+    done
 }
 
 function copy_data
@@ -81,23 +88,27 @@ function copy_data
 	mkdir $DDIR/data/$folder; 
 	cp -p $SDIR/data/$folder/*.{png,jpg} $DDIR/data/$folder
     done
+    mkdir $DDIR/data/gfx/flags25x15
+    cp -p $SDIR/data/gfx/flags25x15/*.png $DDIR/data/gfx/flags25x15
     mkdir $DDIR/data/schemas
     cplffiles data/schemas data/schemas xsd
     cplffiles data/schemas data/schemas xml
     copy_levels
     mkdir $DDIR/data/fonts
     cp -p $SDIR/data/fonts/*.{bmf,png,txt,ttf} $DDIR/data/fonts
-    mkdir $DDIR/data/sound
-    cp -p $SDIR/data/sound/*.{ogg,s3m} $DDIR/data/sound
+    mkdir $DDIR/data/music
+    mkdir $DDIR/data/music/menu
+    cp -p $SDIR/data/music/menu/*.{ogg,s3m} $DDIR/data/music/menu
+    mkdir $DDIR/data/music
+    mkdir $DDIR/data/music/game
+    cp -p $SDIR/data/music/game/*.{ogg,s3m} $DDIR/data/music/game
     mkdir $DDIR/data/soundsets 
     mkdir $DDIR/data/soundsets/enigma 
     cp -p $SDIR/data/soundsets/enigma/*.wav $DDIR/data/soundsets/enigma
-    mkdir $DDIR/data/thumbs
-    cp -p $SDIR/data/thumbs/README $DDIR/data/thumbs
     
     echo "... copying locales:"
     mkdir $DDIR/data/locale
-    for lang in de fr nl it es sv ru hu pt fi; do
+    for lang in de fr nl it es sv ru hu pt fi uk be el pl cs gd hr sk da; do
     	echo "$lang.gmo"
 	mkdir "$DDIR/data/locale/$lang"
 	mkdir "$DDIR/data/locale/$lang/LC_MESSAGES"
@@ -128,6 +139,7 @@ function copy_doc
     cpfile doc/images/flags25x15/*.png	images/flags25x15
     cptext etc/README-SDL.txt 	        README-SDL.txt
     cptext doc/reference/ant_lua.txt 	reference/ant_lua.txt
+    cptext doc/reference/ConceptLuaAPI2.txt 	reference/ConceptLuaAPI2.txt
     cptext doc/reference/sounds.txt	reference/sounds.txt
     cptext doc/reference/soundset.lua   reference/soundset.lua
     cpfile doc/reference/lua2xml	reference/
