@@ -1,12 +1,12 @@
 /*
  * Copyright (C) 2004 Daniel Heck
- * Copyright (C) 2006, 2007 Ronald Lamprecht
+ * Copyright (C) 2006,2007,2008,2009,2010 Ronald Lamprecht
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
  * as published by the Free Software Foundation; either version 2
  * of the License, or (at your option) any later version.
- *  
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
@@ -26,7 +26,9 @@
 #include "gui/HelpMenu.hh"
 #include "main.hh"
 #include "gui/GameMenu.hh"
-#include "sound.hh"
+#include "SoundEngine.hh"
+#include "SoundEffectManager.hh"
+#include "MusicManager.hh"
 #include "player.hh"
 #include "world.hh"
 #include "nls.hh"
@@ -40,6 +42,7 @@
 #include "ecl_sdl.hh"
 
 #include "enet/enet.h"
+#include "enet_ver.hh"
 
 #include <cctype>
 #include <cstring>
@@ -52,7 +55,7 @@ using namespace ecl;
 using namespace std;
 
 #include "client_internal.hh"
-
+
 /* -------------------- Auxiliary functions -------------------- */
 
 namespace
@@ -67,8 +70,7 @@ namespace
     }
 
     /*! Generate the message that is displayed when the level starts. */
-    string displayedLevelInfo (lev::Proxy *level)
-    {
+    std::string displayedLevelInfo (lev::Proxy *level) {
         std::string text;
         std::string tmp;
 
@@ -81,18 +83,18 @@ namespace
             text += _(" by ") + tmp;
         tmp = level->getLocalizedString("subtitle");
         if (!tmp.empty() && tmp != "subtitle")
-           text += string(" - \"")+ tmp + "\""; 
+           text += string(" - \"")+ tmp + "\"";
         tmp = level->getCredits(false);
         if (!tmp.empty())
-            text += string(" - Credits: ")+ tmp; 
+            text += string(" - Credits: ")+ tmp;
         tmp = level->getDedication(false);
         if (!tmp.empty())
-            text += string(" - Dedication: ")+ tmp; 
+            text += string(" - Dedication: ")+ tmp;
         return text;
     }
 }
 
-
+
 /* -------------------- Variables -------------------- */
 
 namespace
@@ -103,36 +105,53 @@ namespace
 
 #define CLIENT client_instance
 
-
 /* -------------------- Client class -------------------- */
 
-Client::Client() 
+Client::Client()
 : m_state (cls_idle),
   m_levelname(),
   m_hunt_against_time(0),
-  m_cheater(false), 
+  m_cheater(false),
   m_user_input()
 {
     m_network_host = 0;
 }
 
-Client::~Client() 
+Client::~Client()
 {
     network_stop();
 }
 
+void Client::init() {
+    std::string command;
+    for (int i = 0; i < 10; i++) {
+        command = app.state->getString(ecl::strf("CommandHistory#%d", i).c_str());
+        if (command.size() > 0)
+            commandHistory.push_back(command);
+        else
+            break;
+    }
+}
+
+void Client::shutdown() {
+    for (int i = 0; i < commandHistory.size(); i++)
+        app.state->setProperty(ecl::strf("CommandHistory#%d", i).c_str(), commandHistory[i].c_str());
+}
+
 bool Client::network_start()
 {
     if (m_network_host)
         return true;
-
     m_network_host = enet_host_create (NULL,
                                        1 /* only allow 1 outgoing connection */,
+#ifdef ENET_VER_EQ_GT_13
+                                       2 /* 2 channels are sufficient */,
+#endif
                                        57600 / 8 /* 56K modem with 56 Kbps downstream bandwidth */,
                                        14400 / 8 /* 56K modem with 14 Kbps upstream bandwidth */);
 
     if (m_network_host == NULL) {
-        fprintf (stderr, 
+        fprintf (stderr,
                  "An error occurred while trying to create an ENet client host.\n");
         return false;
     }
@@ -148,14 +167,18 @@ bool Client::network_start()
     sv_address.port = 12345;
 
     /* Initiate the connection, allocating the two channels 0 and 1. */
-    m_server = enet_host_connect (m_network_host, &sv_address, 2);    
-    
+#ifdef ENET_VER_EQ_GT_13
+    m_server = enet_host_connect (m_network_host, &sv_address, 2, 57600);
+#else
+    m_server = enet_host_connect (m_network_host, &sv_address, 2);
+#endif
+
     if (m_server == NULL) {
-       fprintf (stderr, 
+       fprintf (stderr,
                 "No available peers for initiating an ENet connection.\n");
        return false;
     }
-    
+
     // Wait up to 5 seconds for the connection attempt to succeed.
     ENetEvent event;
     if (enet_host_service (m_network_host, &event, 5000) > 0 &&
@@ -188,7 +211,7 @@ void Client::network_stop ()
 
 /* ---------- Event handling ---------- */
 
-void Client::handle_events() 
+void Client::handle_events()
 {
     SDL_Event e;
     while (SDL_PollEvent(&e)) {
@@ -211,7 +234,7 @@ void Client::handle_events()
         case SDL_ACTIVEEVENT: {
             update_mouse_button_state();
             if (e.active.gain == 0 && !video::IsFullScreen())
-                show_menu();
+                show_menu(false);
             break;
         }
 
@@ -222,18 +245,19 @@ void Client::handle_events()
 
         case SDL_QUIT:
             client::Msg_Command("abort");
+            app.bossKeyPressed = true;
             break;
         }
     }
 }
 
-void Client::update_mouse_button_state() 
+void Client::update_mouse_button_state()
 {
     int b = SDL_GetMouseState(0, 0);
     player::InhibitPickup((b & SDL_BUTTON(1)) || (b & SDL_BUTTON(3)));
 }
 
-void Client::on_mousebutton(SDL_Event &e) 
+void Client::on_mousebutton(SDL_Event &e)
 {
     if (e.button.state == SDL_PRESSED) {
         if (e.button.button == 1) {
@@ -259,116 +283,125 @@ void Client::rotate_inventory (int direction)
     player::RotateInventory(direction);
 }
 
-
 /* -------------------- Console related -------------------- */
 
-class HistoryProxy {
-    static int    instances;
-public:
-    static string content;
-
-    HistoryProxy();
-    ~HistoryProxy() {
-        if (!--instances) app.state->setProperty("CommandHistory", content);
-    }
-};
-
-string HistoryProxy::content;
-int    HistoryProxy::instances = 0;
-
-HistoryProxy::HistoryProxy() {
-    if (!instances++) {
-        content = app.state->getString("CommandHistory");
-        if (content.find(HSEP) == string::npos) content = string(1, HSEP);
-    }
-}
-
-static void user_input_history_append(const string& text, bool at_end = true) {
-    HistoryProxy history;
-    size_t       old_pos = history.content.find(string(1, HSEP)+text+HSEP);
-
-    if (old_pos != string::npos)
-        history.content.erase(old_pos, text.length()+1);
-
-    if (at_end)
-        history.content += text+HSEP;
-    else
-        history.content = string(1, HSEP)+text+history.content;
-}
-
-void Client::process_userinput()
-{
-    if (m_user_input != "") {
-        STATUSBAR->hide_text();
-        string commands = m_user_input;
-
-        user_input_history_append(m_user_input);
-        m_user_input = "";
-
-        size_t sep_pos;
-        while ((sep_pos = commands.find_first_of(';')) != string::npos) {
-            string first_command = commands.substr(0, sep_pos);
-            commands.erase(0, sep_pos+1);
-            server::Msg_Command (first_command);
+void Client::process_userinput() {
+    // no addition of existing commands to history
+    if (consoleIndex == 1) {
+        for (int i = 0; i < commandHistory.size(); i++) {
+            if (newCommand == commandHistory[i]) {
+                // take existing history command instead of new command
+                consoleIndex = i + 2;
+                break;
+            }
         }
-        server::Msg_Command (commands); // last command
     }
+    // resort history with selected command at bottom
+    if (consoleIndex == 1) {
+        if (commandHistory.size() < 10)
+             commandHistory.push_back(std::string(""));
+        for (int i = 8; i >= 0; i--) {
+            if (i < commandHistory.size() - 1)
+                commandHistory[i+1] = commandHistory[i];
+        }
+    } else if (consoleIndex > 1) {
+        newCommand = commandHistory[consoleIndex - 2];
+        for (int i = consoleIndex - 3; i >= 0; i--) {
+            if (i < commandHistory.size())
+                commandHistory[i+1] = commandHistory[i];
+        }
+    } else { // document history or inventory
+        return;
+    }
+    commandHistory[0] = newCommand;
+    newCommand = "";
+    consoleIndex = 0;
+    STATUSBAR->hide_text();
+    server::Msg_Command(commandHistory[0]);
 }
 
-void Client::user_input_append (char c) {
-    m_user_input += c;
-    Msg_ShowText (m_user_input, false);
+void Client::user_input_append(char c) {
+    if (consoleIndex <= 0) {
+        consoleIndex = 1;
+        newCommand = c;
+    } else if (consoleIndex == 1) {
+        newCommand += c;
+    } else {
+        newCommand =  commandHistory[consoleIndex - 2] + c;
+        consoleIndex = 1;
+    }
+    Msg_ShowText(newCommand, false);
 }
 
-void Client::user_input_backspace ()
-{
-    if (!m_user_input.empty()) {
-        m_user_input.erase (m_user_input.size()-1, 1);
-        if (!m_user_input.empty()) {
-            // still not empty
-            Msg_ShowText (m_user_input, false);
+void Client::user_input_backspace() {
+    if (consoleIndex == 1) {
+        newCommand.erase(newCommand.size() - 1, 1);
+        if (!newCommand.empty())
+            Msg_ShowText(newCommand, false);
+        else {
+            consoleIndex = 0;
+            STATUSBAR->hide_text();
+        }
+    } else if (consoleIndex > 1) {
+        newCommand =  commandHistory[consoleIndex - 2];
+        newCommand.erase(newCommand.size() - 1, 1);
+        if (!newCommand.empty()) {
+            consoleIndex = 1;
+            Msg_ShowText(newCommand, false);
         } else {
-            // empty
+            consoleIndex = 0;
             STATUSBAR->hide_text();
         }
     }
 }
-void Client::user_input_previous ()
-{
-    HistoryProxy history;
-    size_t       last_start = history.content.find_last_of(HSEP, history.content.length()-2);
 
-    if (last_start != string::npos) {
-        string prev_input = history.content.substr(last_start+1, history.content.length()-last_start-2);
-        history.content.erase(last_start+1);
-        user_input_history_append(m_user_input, false);
-        m_user_input      = prev_input;
-
-        if (m_user_input.empty())
+void Client::user_input_previous() {
+    if (consoleIndex < 0) {
+        ++consoleIndex;
+        int docIndex = documentHistory.size() + consoleIndex;
+        if (docIndex < documentHistory.size()) {
+            Msg_ShowText(documentHistory[docIndex], true);
+        } else {
+            consoleIndex = 0;
             STATUSBAR->hide_text();
-        else
-            Msg_ShowText (m_user_input, false);
-    }
-}
-void Client::user_input_next ()
-{
-    HistoryProxy history;
-    size_t       first_end = history.content.find_first_of(HSEP, 1);
-
-    if (first_end != string::npos) {
-        string next_input = history.content.substr(1, first_end-1);
-        history.content.erase(0, first_end);
-        user_input_history_append(m_user_input);
-        m_user_input      = next_input;
-
-        if (m_user_input.empty())
-            STATUSBAR->hide_text();
-        else
-            Msg_ShowText (m_user_input, false);
+        }
+    } else if (consoleIndex == 0) {
+        if (newCommand.length() > 0) {
+            consoleIndex = 1;
+            Msg_ShowText(newCommand, false);
+        } else if (commandHistory.size() > 0) {
+            consoleIndex = 2;
+            Msg_ShowText(commandHistory[0], false);
+        }
+    } else if (consoleIndex <= commandHistory.size()) {
+        ++consoleIndex;
+        Msg_ShowText(commandHistory[consoleIndex - 2], false);
+    } else {  // top of history or new command without history
+        consoleIndex = 0;
+        STATUSBAR->hide_text();
     }
 }
 
-void Client::on_keydown(SDL_Event &e) 
+void Client::user_input_next() {
+    if (consoleIndex <= 0) {
+        --consoleIndex;
+        int docIndex = documentHistory.size() + consoleIndex;
+        if (docIndex >= 0) {
+            Msg_ShowText(documentHistory[docIndex], true);
+        } else {
+            consoleIndex = 0;
+            STATUSBAR->hide_text();
+        }
+    } else if (consoleIndex == 1 || (consoleIndex == 2 && newCommand.size() == 0)) {
+        consoleIndex = 0;
+        STATUSBAR->hide_text();
+    } else if (consoleIndex > 1) {
+        --consoleIndex;
+        Msg_ShowText(consoleIndex == 1 ? newCommand : commandHistory[consoleIndex - 2], false);
+    }
+}
+
+void Client::on_keydown(SDL_Event &e)
 {
     SDLKey keysym = e.key.keysym.sym;
     SDLMod keymod = e.key.keysym.mod;
@@ -378,12 +411,10 @@ void Client::on_keydown(SDL_Event &e)
         case SDLK_a:
             server::Msg_Command ("restart");
             break;
-        case SDLK_F3: 
+        case SDLK_F3:
             if (keymod & KMOD_SHIFT) {
                 // force a reload from file
-                lev::Proxy * curProxy = lev::Proxy::loadedLevel();
-                if (curProxy != NULL)
-                    curProxy->release();
+                lev::Proxy::releaseCache();
                 server::Msg_Command ("restart");
             }
         default:
@@ -392,11 +423,12 @@ void Client::on_keydown(SDL_Event &e)
     }
     else if (keymod & KMOD_ALT) {
         switch (keysym) {
-        case SDLK_x: abort(); break;
+        case SDLK_x:
+            abort(); break;
         case SDLK_t:
             if (enigma::WizardMode) {
                 Screen *scr = video::GetScreen();
-                ecl::TintRect(scr->get_surface (), display::GetGameArea(), 
+                ecl::TintRect(scr->get_surface (), display::GetGameArea(),
                              100, 100, 100, 0);
                 scr->update_all();
             }
@@ -417,9 +449,26 @@ void Client::on_keydown(SDL_Event &e)
             break;
         };
     }
+    else if (keymod & KMOD_META) {
+        switch (keysym) {
+            case SDLK_q:    // Mac OS X application quit sequence
+                app.bossKeyPressed = true;
+                abort();
+                break;
+            default:
+                break;
+        };
+    }
     else {
         switch (keysym) {
-        case SDLK_ESCAPE: show_menu(); break;
+        case SDLK_ESCAPE:
+            if (keymod & KMOD_SHIFT) {
+                app.bossKeyPressed = true;
+                abort();
+            } else {
+                show_menu(true);
+            }
+            break;
         case SDLK_LEFT:   set_mousespeed(options::GetMouseSpeed() - 1); break;
         case SDLK_RIGHT:  set_mousespeed(options::GetMouseSpeed() + 1); break;
         case SDLK_TAB:    rotate_inventory(+1); break;
@@ -427,11 +476,11 @@ void Client::on_keydown(SDL_Event &e)
         case SDLK_F2:
             // display hint
             break;
-        case SDLK_F3: 
+        case SDLK_F3:
             if (keymod & KMOD_SHIFT)
                 server::Msg_Command ("restart");
             else
-                server::Msg_Command ("suicide"); 
+                server::Msg_Command ("suicide");
             break;
 
         case SDLK_F4: Msg_AdvanceLevel(lev::ADVANCE_STRICTLY); break;
@@ -439,7 +488,7 @@ void Client::on_keydown(SDL_Event &e)
         case SDLK_F6: Msg_JumpBack(); break;
 
         case SDLK_F10: {
-            lev::Proxy *level = lev::Proxy::loadedLevel();
+            lev::Proxy *level = server::LoadedProxy;
             std::string basename = std::string("screenshots/") +
                     level->getLocalSubstitutionLevelPath();
             std::string fname = basename + ".png";
@@ -475,6 +524,7 @@ static const char *helptext_ingame[] = {
     N_("Left mouse button:"),    N_("Activate/drop leftmost inventory item"),
     N_("Right mouse button:"),      N_("Rotate inventory items"),
     N_("Escape:"),                  N_("Show game menu"),
+    N_("Shift+Escape:"),            N_("Quit game immediately"),
     N_("F1:"),                      N_("Show this help"),
     N_("F3:"),                      N_("Kill current marble"),
     N_("Shift+F3:"),                N_("Restart the current level"),
@@ -488,12 +538,12 @@ static const char *helptext_ingame[] = {
     0
 };
 
-void Client::show_help() 
+void Client::show_help()
 {
     server::Msg_Pause (true);
     video::TempInputGrab grab(false);
 
-    helptext_ingame[15] = app.state->getInt("NextLevelMode") == lev::NEXT_LEVEL_NOT_BEST
+    helptext_ingame[17] = app.state->getInt("NextLevelMode") == lev::NEXT_LEVEL_NOT_BEST
         ? _("Skip to next level for best score hunt")
         : _("Skip to next unsolved level");
 
@@ -508,14 +558,22 @@ void Client::show_help()
     server::Msg_Pause (false);
     game::ResetGameTimer();
 
-    if (app.state->getInt("NextLevelMode") == lev::NEXT_LEVEL_NOT_BEST) 
+    if (app.state->getInt("NextLevelMode") == lev::NEXT_LEVEL_NOT_BEST)
         server::Msg_Command ("restart"); // inhibit cheating
 
 }
 
 
-void Client::show_menu() 
-{
+void Client::show_menu(bool isESC) {
+    if (isESC && server::LastMenuTime != 0.0 && server::LevelTime - server::LastMenuTime < 0.3) {
+        return; // protection against ESC D.o.S. attacks
+    }
+    if (isESC && server::LastMenuTime != 0.0 && server::LevelTime - server::LastMenuTime < 0.35) {
+        server::MenuCount++;
+        if (server::MenuCount > 10)
+            mark_cheater();
+    }
+
     server::Msg_Pause (true);
 
     ecl::Screen *screen = video::GetScreen();
@@ -536,6 +594,8 @@ void Client::show_menu()
     server::Msg_Pause (false);
     game::ResetGameTimer();
 
+    if (isESC)  // protection against ESC D.o.S. attacks
+        server::LastMenuTime = server::LevelTime;
 }
 
 void Client::draw_screen()
@@ -556,7 +616,7 @@ void Client::draw_screen()
         const video::VMInfo *vminfo = video::GetInfo();
         int width = vminfo->width - 120;
         for (unsigned i=0; i<lines.size(); ) {
-            std::string::size_type breakPos = ecl::breakString (f, lines[i], 
+            std::string::size_type breakPos = ecl::breakString (f, lines[i],
                                                                 " ", width);
             f->render(gc, x,  y, lines[i].substr(0,breakPos).c_str());
             y += yskip;
@@ -578,7 +638,7 @@ void Client::draw_screen()
 }
 
 
-std::string Client::init_hunted_time() 
+std::string Client::init_hunted_time()
 {
     std::string hunted;
     m_hunt_against_time = 0;
@@ -587,7 +647,7 @@ std::string Client::init_hunted_time()
         lev::ScoreManager *scm = lev::ScoreManager::instance();
         lev::Proxy *curProxy = ind->getCurrent();
         lev::RatingManager *ratingMgr = lev::RatingManager::instance();
-        
+
         int   difficulty     = app.state->getInt("Difficulty");
         int   wr_time       = ratingMgr->getBestScore(curProxy, difficulty);
         int   best_user_time = scm->getBestUserScore(curProxy, difficulty);
@@ -606,7 +666,7 @@ std::string Client::init_hunted_time()
     return hunted;
 }
 
-void Client::tick (double dtime) 
+void Client::tick (double dtime)
 {
     const double timestep = 0.01; // 10ms
 
@@ -643,17 +703,17 @@ void Client::tick (double dtime)
                     lev::Proxy *curProxy = ind->getCurrent();
                     lev::RatingManager *ratingMgr = lev::RatingManager::instance();
                     int    difficulty     = app.state->getInt("Difficulty");
-                    int    wr_time       = ratingMgr->getBestScore(curProxy, difficulty);
+                    int    wr_time        = ratingMgr->getBestScore(curProxy, difficulty);
                     int    best_user_time = scm->getBestUserScore(curProxy, difficulty);
                     string message;
 
                     if (wr_time>0 && (best_user_time<0 || best_user_time>wr_time)) {
-                        message = string(_("Too slow for ")) + 
+                        message = string(_("Too slow for ")) +
                             ratingMgr->getBestScoreHolder(curProxy, difficulty) +
-                            ".. [Ctrl-A]";
+                            "... Ctrl-A";
                     }
                     else {
-                        message = string(_("You are slow today.. [Ctrl-A]"));
+                        message = string(_("You are slow today ... Ctrl-A"));
                     }
 
                     client::Msg_PlaySound("shatter", 1.0);
@@ -689,15 +749,17 @@ void Client::tick (double dtime)
         break;
     case cls_abort:
         break;
-    case cls_error: 
+    case cls_error:
         {
             SDL_Event e;
             while (SDL_PollEvent(&e)) {
                 switch (e.type) {
-                case SDL_KEYDOWN:
-                case SDL_QUIT:
-                    client::Msg_Command("abort");
-                    break;
+                    case SDL_QUIT:
+                        app.bossKeyPressed = true;
+                        // fall through
+                    case SDL_KEYDOWN:
+                        client::Msg_Command("abort");
+                        break;
                 }
             }
         }
@@ -705,7 +767,7 @@ void Client::tick (double dtime)
     }
 }
 
-void Client::level_finished() 
+void Client::level_finished()
 {
     lev::Index *ind = lev::Index::getCurrentIndex();
     lev::ScoreManager *scm = lev::ScoreManager::instance();
@@ -714,7 +776,6 @@ void Client::level_finished()
     int    difficulty     = app.state->getInt("Difficulty");
     int    wr_time       = ratingMgr->getBestScore(curProxy, difficulty);
     int    best_user_time = scm->getBestUserScore(curProxy, difficulty);
-    string par_name       = ratingMgr->getBestScoreHolder(curProxy, difficulty);
     int par_time   = ratingMgr->getParScore(curProxy, difficulty);
 
     int    level_time     = round_nearest<int> (m_total_game_time);
@@ -722,6 +783,10 @@ void Client::level_finished()
     string      text;
     bool        timehunt_restart = false;
 
+    std::string par_name  = ratingMgr->getBestScoreHolder(curProxy, difficulty);
+    for (int cut = 2; par_name.length() > 55; cut++)
+        par_name  = ratingMgr->getBestScoreHolder(curProxy, difficulty, cut);
+
     if (wr_time > 0) {
         if (best_user_time<0 || best_user_time>wr_time) {
             if (level_time == wr_time)
@@ -736,13 +801,16 @@ void Client::level_finished()
             if (app.state->getInt("NextLevelMode") == lev::NEXT_LEVEL_NOT_BEST)
                 timehunt_restart = true; // when hunting yourself: Equal is too slow
         }
-        else if (level_time<best_user_time)
+        else if (level_time<best_user_time) {
             if (par_time >= 0 && level_time <= par_time)
                 text = _("New personal record - better than par!");
-            else if (par_time >= 0)
-                text = _("New personal record, but over par!");
+            // Uncomment the following lines to show the "but over par!"-part.
+            // (This has been criticised as demoralizing.)
+            //else if (par_time >= 0)
+            //    text = _("New personal record, but over par!");
             else
                 text = _("New personal record!");
+        }
     }
 
     if (app.state->getInt("NextLevelMode") == lev::NEXT_LEVEL_NOT_BEST &&
@@ -758,8 +826,8 @@ void Client::level_finished()
             else {
                 text = "";
             }
-            text += strf("%d:%02d ", static_cast<int> (behind/60)%100, behind%60);
-            if (with_par) 
+            text += ecl::timeformat(behind);
+            if (with_par)
                 text += _("behind world record.");
             else
                 text += _("behind your record.");
@@ -771,15 +839,17 @@ void Client::level_finished()
     if (text.length() == 0) {
         if (par_time >= 0 && level_time <= par_time)
             text = _("Level finished - better than par!");
-        else if (par_time >= 0)
-            text = _("Level finished, but over par!");
+        // Uncomment the following lines to show the "but over par!"-part.
+        // (This has been criticised as demoralizing.)
+        //else if (par_time >= 0)
+        //    text = _("Level finished, but over par!");
         else
             text = _("Level finished!");
     }
     if (m_cheater)
         text += _(" Cheater!");
 
-    Msg_ShowText (text, false);
+    Msg_ShowText(text, false);
 
     if (!m_cheater) {
         scm->updateUserScore(curProxy, difficulty, level_time);
@@ -791,14 +861,10 @@ void Client::level_finished()
 
     if (timehunt_restart)
         server::Msg_Command("restart");
-    else 
+    else
         m_state = cls_finished;
 }
 
-#ifdef _MSC_VER
-#define snprintf _snprintf
-#endif
-
 void Client::level_loaded(bool isRestart)
 {
     lev::Index *ind = lev::Index::getCurrentIndex();
@@ -809,34 +875,27 @@ void Client::level_loaded(bool isRestart)
     video::SetCaption(ecl::strf(_("Enigma pack %s - level #%d: %s"), ind->getName().c_str(),
             ind->getCurrentLevel(), curProxy->getTitle().c_str()).c_str());
 
-    string hunted = init_hunted_time();   // sets m_hunt_against_time (used below)
+    std::string hunted = init_hunted_time();   // sets m_hunt_against_time (used below)
+    documentHistory.clear();
+    consoleIndex = 0;
 
     // show level information (name, author, etc.)
-    {
-        string displayed_info = "";
-        if (m_hunt_against_time>0) {
-            if (hunted == "you")
-                displayed_info = _("Your record: ");
-            else
-                displayed_info = _("World record to beat: ");
-            displayed_info += ecl::strf("%d:%02d", (m_hunt_against_time/60)%100, 
-                    m_hunt_against_time%60);
+    std::string displayed_info = "";
+    if (m_hunt_against_time>0) {
+        if (hunted == "you")
+            displayed_info = _("Your record: ");
+        else
+            displayed_info = _("World record to beat: ");
+        displayed_info += ecl::timeformat(m_hunt_against_time);
 //+ _(" by ") +hunted;
 // makes the string too long in many levels
-            Msg_ShowText (displayed_info, true, 4.0);
-        }
-        else {
-            displayed_info = displayedLevelInfo(curProxy);
-            Msg_ShowText (displayed_info, true, 2.0);
-        }
+        Msg_ShowDocument(displayed_info, true, 4.0);
+    } else {
+        displayed_info = displayedLevelInfo(curProxy);
+        Msg_ShowDocument(displayed_info, true, 2.0);
     }
 
-    sound::FadeoutMusic();
-    if (options::GetBool("InGameMusic")) {
-        sound::PlayMusic (options::GetString("LevelMusicFile"));
-    } else {
-        sound::StopMusic();
-    }
+    sound::StartLevelMusic();
 
     // start screen transition
     GC gc(video::BackBuffer());
@@ -860,18 +919,33 @@ void Client::handle_message (Message *m) { // @@@ unused
     }
 }
 
-void Client::error (const string &text) 
+void Client::error (const string &text)
 {
     m_error_message = text;
     m_state = cls_error;
     draw_screen();
 }
 
-
+void Client::registerDocument(std::string text) {
+    documentHistory.push_back(text);
+    consoleIndex = -1;
+}
+
+void Client::finishedText() {
+    consoleIndex = 0;
+}
+
 /* -------------------- Functions -------------------- */
 
+void client::ClientInit() {
+    CLIENT.init();
+}
 
-bool client::NetworkStart() 
+void client::ClientShutdown() {
+    CLIENT.shutdown();
+}
+
+bool client::NetworkStart()
 {
     return CLIENT.network_start();
 }
@@ -933,7 +1007,7 @@ void client::Msg_Command(const string& cmd) {
     }
 }
 
-void client::Msg_PlayerPosition (unsigned iplayer, const V2 &pos) 
+void client::Msg_PlayerPosition (unsigned iplayer, const V2 &pos)
 {
     if (iplayer == (unsigned)player::CurrentPlayer()) {
         sound::SetListenerPosition (pos);
@@ -941,7 +1015,7 @@ void client::Msg_PlayerPosition (unsigned iplayer, const V2 &pos)
     }
 }
 
-void client::Msg_PlaySound (const std::string &wavfile, 
+void client::Msg_PlaySound (const std::string &wavfile,
                             const ecl::V2 &pos,
                             double relative_volume)
 {
@@ -954,16 +1028,23 @@ void client::Msg_PlaySound (const std::string &wavfile, double relative_volume)
 }
 
 void client::Msg_Sparkle (const ecl::V2 &pos) {
-    display::AddEffect (pos, "ring-anim");
+    display::AddEffect (pos, "ring-anim", true);
 }
 
 
-void client::Msg_ShowText
-(const std::string &text, bool scrolling, double duration)
-{
+void client::Msg_ShowText(const std::string &text, bool scrolling, double duration) {
     STATUSBAR->show_text (text, scrolling, duration);
 }
 
+void client::Msg_ShowDocument(const std::string &text, bool scrolling, double duration) {
+    CLIENT.registerDocument(text);
+    Msg_ShowText(text, scrolling, duration);
+}
+
+void client::Msg_FinishedText() {
+    CLIENT.finishedText();
+}
+
 void client::Msg_Error (const std::string &text)
 {
     CLIENT.error (text);
