@@ -19,7 +19,7 @@
 
 #include "gui/widgets.hh"
 #include "enigma.hh"
-#include "sound.hh"
+#include "SoundEffectManager.hh"
 #include "video.hh"
 #include "options.hh"
 #include "nls.hh"
@@ -98,7 +98,7 @@ void Image::draw (ecl::GC &gc, const ecl::Rect &/*r*/) {
     }
 }
 
-
+
 /* -------------------- AreaManager -------------------- */
 
 AreaManager::AreaManager(Container *c)
@@ -180,7 +180,7 @@ void Container::remove_child (Widget *w) {
 }
 
 void Container::exchange_child (Widget *oldChild, Widget *newChild) {
-    for (int i = 0; i < m_widgets.size(); i++) {
+    for (unsigned int i = 0; i < m_widgets.size(); i++) {
         if (m_widgets[i] == oldChild) {
             m_widgets[i] = newChild;
             newChild->set_parent(this);
@@ -199,8 +199,8 @@ void Container::draw (ecl::GC& gc, const ecl::Rect &r) {
 }
 
 void Container::draw_all() {
-        invalidate_all();
-        refresh();
+    invalidate_all();
+    refresh();
 }
     
 void Container::reconfigure_child (Widget *)
@@ -291,7 +291,19 @@ ecl::Rect Container::boundingbox() {
         return get_area();
 }
 
-
+void Container::set_key_focus(Widget *newfocus) {
+    if (get_parent() != NULL) {
+        get_parent()->set_key_focus(newfocus);
+    }
+}
+
+bool Container::is_key_focus(Widget *focus) {
+    if (get_parent() != NULL) {
+        return get_parent()->is_key_focus(focus);
+    } else
+        return false;
+}
+
 /* -------------------- List -------------------- */
 
 List::List (int spacing) 
@@ -455,7 +467,6 @@ bool HList::fits() {
     return targetw >= naturalw;
 }
 
-
 /* -------------------- VList -------------------- */
 
 void VList::recalc()
@@ -505,7 +516,6 @@ bool VList::fits() {
     return targeth >= naturalh;
 }
 
-
 /* -------------------- Label -------------------- */
 
 Label::Label (const std::string &text,
@@ -584,7 +594,6 @@ void Label::naturalsize (int &w, int &h) const
     w = m_font->get_width (m_text == "" ? "" : get_text().c_str());
 }
 
-
 /* -------------------- UntranslatedLabel -------------------- */
 
 UntranslatedLabel::UntranslatedLabel (const std::string &text,
@@ -598,7 +607,9 @@ string UntranslatedLabel::get_text() const {
 
 /* -------------------- Button -------------------- */
 
-Button::Button() : m_activep (false), highlight (false) {
+Button::Button()
+: m_activep (false), highlight (false)
+{
 }
 
 void Button::activate() 
@@ -624,7 +635,7 @@ bool Button::isHighlight() {
 void Button::draw(ecl::GC &gc, const ecl::Rect &r) {
     const int borderw = 4;
 
-    ecl::Surface *s = enigma::GetImage (m_activep ? "buttonhl" : "button");
+    ecl::Surface *s = enigma::GetImage(m_activep ? "buttonhl" : "button");
 
     if (s) {                    // Ugly, but hey, it works
         Rect srcrect (0,0,borderw, borderw);
@@ -690,7 +701,6 @@ void Button::draw(ecl::GC &gc, const ecl::Rect &r) {
     }
 }
 
-
 /* -------------------- PushButton -------------------- */
 
 PushButton::PushButton() : m_pressedp (false) {
@@ -699,30 +709,30 @@ PushButton::PushButton() : m_pressedp (false) {
 bool PushButton::on_event(const SDL_Event &e) {
     Widget::on_event(e);
     bool was_pressed = m_pressedp;
+    bool handeled = false;
 
     switch (e.type) {
-    case SDL_KEYDOWN:
-        if (e.key.keysym.sym != SDLK_RETURN &&
-            e.key.keysym.sym != SDLK_SPACE) break;
-        // fall-through
-    case SDL_MOUSEBUTTONDOWN:
-        m_pressedp = true;
-        break;
-
-    case SDL_KEYUP:
-        if (e.key.keysym.sym != SDLK_RETURN &&
-            e.key.keysym.sym != SDLK_SPACE &&
-            e.key.keysym.sym != SDLK_PAGEDOWN &&
-            e.key.keysym.sym != SDLK_PAGEUP) break;
-        lastUpSym = e.key.keysym.sym;
-        lastUpBotton = 0;
-        m_pressedp = false;
-        break;
-    case SDL_MOUSEBUTTONUP:
-        lastUpSym = SDLK_UNKNOWN;
-        lastUpBotton = e.button.button;
-        m_pressedp = false;
-        break;
+        case SDL_KEYDOWN:
+            if (e.key.keysym.sym != SDLK_RETURN) break;
+            // fall-through
+        case SDL_MOUSEBUTTONDOWN:
+            m_pressedp = true;
+            handeled = true;
+            break;
+    
+        case SDL_KEYUP:
+            if (e.key.keysym.sym != SDLK_RETURN) break;
+            lastUpSym = e.key.keysym.sym;
+            lastUpBotton = 0;
+            m_pressedp = false;
+            handeled = true;
+            break;
+        case SDL_MOUSEBUTTONUP:
+            lastUpSym = SDLK_UNKNOWN;
+            lastUpBotton = e.button.button;
+            m_pressedp = false;
+            handeled = true;
+            break;
     }
 
     bool changed = (was_pressed != m_pressedp);
@@ -735,7 +745,7 @@ bool PushButton::on_event(const SDL_Event &e) {
         }
     }
 
-    return changed;
+    return handeled;
 }
 
 void PushButton::deactivate() {
@@ -757,17 +767,12 @@ Uint8 PushButton::getLastUpButton() {
 bool PushButton::soundOk() {
     return true;
 }
-
+
 /* -------------------- TextButton -------------------- */
 
-ecl::Font *TextButton::menufont = 0;
-ecl::Font *TextButton::menufont_pressed = 0;
-
 TextButton::TextButton(ActionListener *al) {
-    if (menufont == 0) {
-        menufont = enigma::GetFont("menufont");
-        menufont_pressed = enigma::GetFont("menufontsel");
-    }
+    menufont = enigma::GetFont("menufont");
+    menufont_pressed = enigma::GetFont("menufontsel");
     set_listener(al);
 }
 
@@ -783,7 +788,7 @@ void TextButton::draw(ecl::GC &gc, const ecl::Rect &r) {
     f->render (gc, x, y, text.c_str());
 }
 
-
+
 /* -------------------- StaticTextButton -------------------- */
 
 StaticTextButton::StaticTextButton(const string &t, ActionListener *al)
@@ -816,7 +821,7 @@ string UntranslatedStaticTextButton::get_text() const {
     return StaticTextButton::text;
 }
 
-
+
 /* -------------------- Buttons for Options -------------------- */
 
 BoolOptionButton::BoolOptionButton(const char    *option_name,
@@ -843,8 +848,7 @@ void BoolOptionButton::on_action(Widget *) {
 string BoolOptionButton::get_text() const {
     return enigma_options::GetBool(optionName) ? _(trueText.c_str()) : _(falseText.c_str());
 }
-    
-
+
 /* -------------------- ValueButton -------------------- */
 
 ValueButton::ValueButton(int min_value_, int max_value_)
@@ -864,7 +868,14 @@ void ValueButton::init() {
 
 bool ValueButton::inc_value(int offset) {
     int old_value = get_value();
-    return update_value(old_value, old_value+offset);
+    int new_value = old_value + offset;
+    if(new_value > max_value) {
+        new_value = min_value;
+    }
+    if(new_value < min_value) {
+        new_value = max_value;
+    }
+    return update_value(old_value, new_value);
 }
 
 string ValueButton::get_text() const {
@@ -885,13 +896,16 @@ bool ValueButton::update_value(int old_value, int new_value) {
 void ValueButton::on_action(Widget *) {
     int incr = 1;
     bool stop = false;
-    if (getLastUpSym() == SDLK_PAGEDOWN || getLastUpButton() == SDL_BUTTON_RIGHT ||
-            getLastUpButton() == 5) {  // wheel down
+    if (getLastUpSym() == SDLK_PAGEDOWN || 
+        getLastUpButton() == SDL_BUTTON_RIGHT ||
+        getLastUpButton() == 5) {  // wheel down
         incr = -1;
     }
-    if (getLastUpSym() == SDLK_PAGEDOWN || getLastUpSym() == SDLK_PAGEUP ||
-            getLastUpButton() == SDL_BUTTON_RIGHT ||
-            getLastUpButton() ==  4 || getLastUpButton() == 5) {
+    if (getLastUpSym() == SDLK_PAGEDOWN || 
+        getLastUpSym() == SDLK_PAGEUP ||
+        getLastUpButton() == SDL_BUTTON_RIGHT ||
+        getLastUpButton() == SDL_BUTTON_LEFT  ||
+        getLastUpButton() ==  4 || getLastUpButton() == 5) {
         stop = true;
     }
     if (inc_value(incr)) {
@@ -912,7 +926,6 @@ void ValueButton::on_action(Widget *) {
 bool ValueButton::soundOk() {
     return false;
 }
-
 
 /* -------------------- ImageButton -------------------- */
 
@@ -941,3 +954,40 @@ void ImageButton::draw(ecl::GC &gc, const ecl::Rect &r) {
         blit(gc, x, y, s);
     }
 }
+
+/* -------------------- BorderlessImageButton -------------------- */
+
+BorderlessImageButton::BorderlessImageButton(const string &unselected,
+    const string &selected, const string &mouseover, bool isSelected, ActionListener *al)
+: fname_sel(selected), fname_unsel(unselected), fname_mouse(mouseover), state (isSelected)
+{
+    set_listener(al);
+}
+
+void BorderlessImageButton::set_images(const string &unselected,
+                                       const string &selected, const string &mouseover) {
+    fname_sel = selected;
+    fname_unsel = unselected;
+    fname_mouse = mouseover;
+}
+
+void BorderlessImageButton::draw(ecl::GC &gc, const ecl::Rect &r) {
+    string &fname = m_activep ? fname_mouse : (state ? fname_sel : fname_unsel);
+
+    if (Surface *s = enigma::GetImage(fname.c_str())) {
+        int w=s->width();
+        int h=s->height();
+        int x = get_x() + (get_w()-w)/2;
+        int y = get_y() + (get_h()-h)/2;
+        blit(gc, x, y, s);
+    }
+}
+
+void BorderlessImageButton::setState(bool isSelected) {
+    state = isSelected;
+    invalidate();
+}
+
+bool BorderlessImageButton::getState() const {
+    return state;
+}
