@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2002,2003,2004,2005 Daniel Heck
+ * Copyright (C) 2006,2007,2008,2009 Ronald Lamprecht
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -16,7 +17,6 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
  */
-
 #include "main.hh"
 #include "display.hh"
 #include "lua.hh"
@@ -25,7 +25,9 @@
 #include "gui/LevelPreviewCache.hh"
 #include "options.hh"
 #include "oxyd.hh"
-#include "sound.hh"
+#include "SoundEngine.hh"
+#include "SoundEffectManager.hh"
+#include "MusicManager.hh"
 #include "video.hh"
 #include "ecl_argp.hh"
 #include "ecl_system.hh"
@@ -33,6 +35,7 @@
 #include "world.hh"
 #include "nls.hh"
 #include "LocalToXML.hh"
+#include "ObjectValidator.hh"
 #include "PreferenceManager.hh"
 #include "Utf8ToXML.hh"
 #include "XMLtoUtf8.hh"
@@ -45,7 +48,6 @@
 #include "enet/enet.h"
 
 #include <locale.h>
-
 #include <cstdio>
 #include <cstdlib>
 #include <fstream>
@@ -56,17 +58,23 @@
 #include <xercesc/util/XMLUniDefs.hpp>
 #include <xercesc/util/PlatformUtils.hpp>
 #include <xercesc/util/XercesVersion.hpp>
-
+#include "SDL_main.h"
+#include <SDL_image.h>
+#include <SDL_mixer.h>
+#include <SDL_ttf.h>
 
 #ifdef MACOSX
 // for search paths
 #include "NSSystemDirectories.h"
+#include <Carbon/Carbon.h>
+#include <CoreFoundation/CoreFoundation.h>
+#include <CoreServices/CoreServices.h>
 #endif
 
 using namespace std;
 using namespace ecl;
 using namespace enigma;
-XERCES_CPP_NAMESPACE_USE 
+XERCES_CPP_NAMESPACE_USE
 
 #ifdef WIN32
 // LoadImage is a Syscall on Windows, which gets defined to LoadImageA
@@ -80,24 +88,25 @@ namespace
     class Nulbuf : public std::streambuf {};
     Nulbuf* nullbuffer = new Nulbuf;
 }
-
+
 /* -------------------- Variables -------------------- */
 
 namespace enigma
 {
     Application app;
-    
+
     bool noAssert = true;      // block expensive assert evaluations by default
 
     bool   WizardMode        = false;
 
     //! If true, do not ``grab'' the mouse and keyboard
     bool   Nograb            = false;
+    bool Robinson = false;   // do not connect to internet if true
 }
 
 /*! The stream object that is used for logging messages.  As defined
   here, it is not connected to any file or buffer.
-  
+
   (Note: I think writing to a stream without a streambuffer *should*
   be a no-op, but it leads to crashes with g++ 2.95.  to circumvent
   this, Log is initialized with a dummy streambuf in init(). ) */
@@ -111,17 +120,22 @@ static std::fstream logfile;
 
 static void usage()
 {
-    printf("Usage: %s [options] [level files]\n\n"
+    printf("Usage: %s [OPTION]... [levelfile.xml]...\n\n"
            "Available options :\n\n"
-           "    --nosound       Disable music and sound\n"
-           "    --nomusic       Disable music\n"
-           "    --window -w     Run in a window; do not enter fullscreen mode\n"
-           "    --help -h       Show this help\n"
-           "    --version       Print the executable's version number\n"
-           "    --nograb        Do not use exclusive mouse/keyboard access\n"
-           "    --data -d path  Load data from additional directory\n"
-           "    --lang -l lang  Set game language\n"
-           "    --pref -p file  Use filename or dirname for preferences\n"
+           "    --assert       Evaluate all debugging assertions\n"
+           "    --data -d path Load data from additional directory\n"
+           "    --help -h      Show this help\n"
+           "    --lang -l lang Set game language\n"
+           "    --log          Turn on logging to the standard output\n"
+           "    --nograb       Do not use exclusive mouse/keyboard access\n"
+           "    --nomusic      Disable music\n"
+           "    --nosound      Disable music and sound effects\n"
+           "    --pref -p file Use filename or dirname for preferences\n"
+           "    --redirect     Redirect stdout/stderr to files on user path\n"
+           "    --robinson     avoid connections to the internet\n"
+           "    --showfps      Show the framerate (FPS) during the Game\n"
+           "    --version      Print the executable's version number\n"
+           "    --window -w    Run in a window; do not enter fullscreen mode\n"
            "\n",
            app.progCallPath.c_str()
            );
@@ -136,7 +150,7 @@ namespace
 
         // Variables.
         bool nosound, nomusic, show_help, show_version, do_log, do_assert, force_window;
-        bool dumpinfo, makepreview;
+        bool dumpinfo, makepreview, show_fps, redirect;
         string gamename;
         string datapath;
         string preffilename;
@@ -155,14 +169,13 @@ namespace
 
         void on_option (int id, const string &param);
         void on_argument (const string &arg);
-
     };
 }
 
 AP::AP() : ArgParser (app.args.begin(), app.args.end())
 {
     nosound  = nomusic = show_help = show_version = do_log = do_assert = force_window = false;
-    dumpinfo = makepreview = false;
+    dumpinfo = makepreview = show_fps = redirect = false;
     gamename = "";
     datapath = "";
     preffilename = PREFFILENAME;
@@ -177,6 +190,9 @@ AP::AP() : ArgParser (app.args.begin(), app.args.end())
     def (&do_assert,            "assert");
     def (&dumpinfo,             "dumpinfo");
     def (&makepreview,          "makepreview");
+    def (&show_fps,             "showfps");
+    def (&redirect,             "redirect");
+    def (&Robinson,             "robinson");
     def (&force_window,         "window", 'w');
     def (OPT_GAME,              "game", true);
     def (OPT_DATA,              "data", 'd', true);
@@ -184,11 +200,11 @@ AP::AP() : ArgParser (app.args.begin(), app.args.end())
     def (OPT_PREF,              "pref", 'p', true);
 }
 
-void AP::on_option (int id, const string &param) 
+void AP::on_option (int id, const string &param)
 {
     switch (id) {
     case OPT_GAME:
-        gamename = param; 
+        gamename = param;
         break;
     case OPT_DATA:
 // we should be able to add several paths -- file.cc has does not yet support this
@@ -206,7 +222,7 @@ void AP::on_option (int id, const string &param)
     }
 }
 
-void AP::on_argument (const string &arg) 
+void AP::on_argument (const string &arg)
 {
     levelnames.push_back (arg);
 }
@@ -222,15 +238,34 @@ void AP::on_argument (const string &arg)
 
 Application::Application() : wizard_mode (false), nograb (false), language (""),
         defaultLanguage (""), argumentLanguage (""), errorInit (false),
-        isMakePreviews (false) {
+        isMakePreviews (false), bossKeyPressed (false) {
 }
 
 
-void Application::init(int argc, char **argv) 
+void Application::init(int argc, char **argv)
 {
+    sscanf(PACKAGE_VERSION, "%4lf", &enigmaVersion);
+
     progCallPath = argv[0];
-    copy(argv+1, argv+argc, back_inserter(args));
-    
+#if MACOSX
+    CFBundleRef mainBundle = CFBundleGetMainBundle();
+    CFURLRef cfurlmain = CFBundleCopyExecutableURL(mainBundle);
+    CFStringRef cffileStr = CFURLCopyFileSystemPath(cfurlmain, kCFURLPOSIXPathStyle);
+    CFIndex cfmax = CFStringGetMaximumSizeOfFileSystemRepresentation(cffileStr);
+    char localbuffer[cfmax];
+    if (CFStringGetFileSystemRepresentation(cffileStr, localbuffer, cfmax)) {
+      progCallPath = localbuffer; // error skips this and defaults to argv[0] which works for most purposes
+    }
+    CFRelease(cfurlmain);
+    CFRelease(cffileStr);
+#endif
+    for (int i=1; i<argc; i++) {
+        if (0 == strcmp(argv[i], "-NSDocumentRevisionsDebugMode"))
+            i++;
+        else
+            args.push_back(argv[i]);
+    }
+
     // parse commandline arguments -- needs args
     AP ap;
     ap.parse();
@@ -242,7 +277,7 @@ void Application::init(int argc, char **argv)
         if (ap.show_help) usage();
         exit(0);
     }
-    
+
     //
     if (ap.makepreview) {
         ap.force_window = true;
@@ -251,23 +286,55 @@ void Application::init(int argc, char **argv)
         isMakePreviews = true;
     }
 
-    // initialize logfile -- needs ap
-    if (ap.do_log) 
-        enigma::Log.rdbuf(cout.rdbuf());
-    else
-        enigma::Log.rdbuf(::nullbuffer);
-
     // initialize assertion stop flag
     if (ap.do_assert)
         enigma::noAssert = false;
 
-    // initialize system datapaths -- needs ap, log
+    // initialize system datapaths -- needs ap
     systemCmdDataPath = ap.datapath;
     initSysDatapaths(ap.preffilename);
 
+    // redirect stdout, stderr
+#ifdef WIN32
+    ap.redirect = true;
+#endif
+    if (ap.redirect) {
+        FILE *newfp;
+        newfp = std::freopen((userStdPath + "/Output.log").c_str(), "w", stdout);
+        if ( newfp == NULL ) {  // This happens on NT
+            newfp = fopen((userStdPath + "/Output.log").c_str(), "w");
+            if (newfp) {  // in case stdout is a macro
+                *stdout = *newfp;
+            }
+        }
+        setvbuf(stdout, NULL, _IOLBF, BUFSIZ);   // Line buffered
+
+        newfp = std::freopen((userStdPath + "/Error.log").c_str(), "w", stderr);
+        if ( newfp == NULL ) {  // This happens on NT
+            newfp = fopen((userStdPath + "/Error.log").c_str(), "w");
+            if (newfp) {  // in case stderr is a macro
+                *stderr = *newfp;
+            }
+        }
+        setbuf(stderr, NULL);   // No buffering
+    }
+
+    // initialize logfile -- needs ap
+    if (ap.do_log)
+        enigma::Log.rdbuf(cout.rdbuf());
+    else
+        enigma::Log.rdbuf(::nullbuffer);
+
+    // postponed system datapath logs
+    Log << "Enigma " << getVersionInfo() << "\n";
+    Log << "systemFS = \"" << systemFS->getDataPath() << "\"\n";
+    Log << "docPath = \"" << docPath << "\"\n";
+    Log << "l10nPath = \"" << l10nPath << "\"\n";
+    Log << "prefPath = \"" << prefPath << "\"\n";
+
     // initialize XML -- needs log, datapaths
     initXerces();
-    
+
     // initialize LUA - Run initialization scripts
     lua_State *L = lua::GlobalState();
     lua::CheckedDoFile(L, app.systemFS, "startup.lua");
@@ -275,31 +342,42 @@ void Application::init(int argc, char **argv)
     // initialize preferences -- needs LUA, XML
     if (!options::Load()) {
         fprintf(stderr, _("Error in configuration file.\n"));
-      	fprintf(stderr, lua::LastError (lua::GlobalState()).c_str());
-    }     
+      	fprintf(stderr, "%s\n", lua::LastError (lua::GlobalState()).c_str());
+    }
     prefs = PreferenceManager::instance();
-    
+
     if (ap.force_window) {
-        options::SetOption("FullScreen", false);
+        app.prefs->setProperty("FullScreen", false);
     }
     if (isMakePreviews) {
-        options::SetOption("VideoMode", 0);
+        app.prefs->setProperty("VideoModesFullscreen", "-0-");
+        app.prefs->setProperty("VideoModesWindow", "-0-");     // we will not save the prefs!
     }
 
     // initialize user data paths -- needs preferences, system datapaths
     initUserDatapaths();
-    
+
     // set message language
     init_i18n();
-    
+
+    // initialize application state
+    state = StateManager::instance();
+
+    // ----- Object declarations
+    ObjectValidator::didInitXML();  // enable initialization
+    ObjectValidator::instance();    // force early initialization (an optional statement)
+
     // ----- Initialize object repositories
-    world::Init();
+    InitWorld();
     if (ap.dumpinfo) {
-        world::DumpObjectInfo();
+        DumpObjectInfo();
         exit(0);
     }
 
     // ----- Initialize SDL library
+#ifdef WIN32
+//    SDL_putenv("SDL_VIDEODRIVER=directx");  //needed for SDL 1.2.12 that crashes on SetGamma on default GDI driver
+#endif
     int sdl_flags = SDL_INIT_VIDEO;
     if (enigma::WizardMode)
         sdl_flags |= SDL_INIT_NOPARACHUTE;
@@ -307,8 +385,27 @@ void Application::init(int argc, char **argv)
         fprintf(stderr, "Couldn't initialize SDL: %s\n", SDL_GetError());
         exit(1);
     }
-    atexit(SDL_Quit);
+    std::atexit(SDL_Quit);
     SDL_EnableUNICODE(1);
+    const SDL_version* vi = SDL_Linked_Version();
+    Log << ecl::strf("SDL Version: %u.%u.%u\n", vi->major, vi->minor, vi->patch);
+
+    vi = TTF_Linked_Version();
+    Log <<  ecl::strf("SDL_ttf Version: %u.%u.%u\n", vi->major, vi->minor, vi->patch);
+    if(TTF_Init() == -1) {
+        fprintf(stderr, "Couldn't initialize SDL_ttf: %s\n", TTF_GetError());
+        exit(1);
+    }
+
+    vi = IMG_Linked_Version();
+    Log <<  ecl::strf("SDL_image Version: %u.%u.%u\n", vi->major, vi->minor, vi->patch);
+#ifdef SDL_IMG_INIT
+    int img_flags = IMG_INIT_PNG | IMG_INIT_JPG;
+    if (IMG_Init(img_flags) & (img_flags != img_flags)) {
+        fprintf(stderr, "Couldn't initialize SDL_image: %s\n", IMG_GetError());
+        exit(1);
+    }
+#endif
 
     // ----- Initialize video subsystem
     video::Init();
@@ -320,25 +417,24 @@ void Application::init(int argc, char **argv)
 
 
     // ----- Initialize sound subsystem
-    if (ap.nosound)
-        sound::DisableSound();
-    else if (ap.nomusic)
-        sound::DisableMusic();
-    sound::Init();
+    sound::Init(!ap.nomusic, !ap.nosound);
     lua::CheckedDoFile (L, app.systemFS, "sound-defaults.lua");
     lua::DoSubfolderfile (L, "soundsets", "soundset.lua");
 
-    // ----- Initialize network layer
-    if (enet_initialize () != 0) {
+    // ----- Initialize TCP network layer
+    if (!InitCurl()) {
+        fprintf (stderr, "An error occurred while initializing Curl.\n");
+        exit (1);
+    }
+
+    // ----- Initialize UDP network layer
+    if (enet_initialize() != 0) {
         fprintf (stderr, "An error occurred while initializing ENet.\n");
         exit (1);
     }
 
     // ----- Load models
-    display::Init();
-
-    // initialize application state
-    state = StateManager::instance();
+    display::Init(ap.show_fps);
 
     // ----- Load level packs -- needs state
     lev::Index::initGroups();
@@ -363,22 +459,27 @@ void Application::init(int argc, char **argv)
     // ----- Initialize sound tables -- needs sound, oxyd, video (error messages!)
     sound::InitSoundSets();
 
+    // ----- Initialize music -- needs application state
+    sound::InitMusic();
+
 #if MACOSX
     updateMac1_00();
 #endif
 
     // initialize random
-    enigma::Randomize();
-    
+    enigma::Randomize(false);
+    enigma::Randomize(true);
+
     //
     if (isMakePreviews) {
+        app.state->setProperty("Difficulty", DIFFICULTY_HARD); // will not be saved
         std::set<lev::Proxy *> proxies = lev::Proxy::getProxies();
         int size = proxies.size();
         std::set<lev::Proxy *>::iterator it;
-        std::string message = ecl::strf("Make %d previews on system path '%s'",
+        std::string message = ecl::strf("Make 3 x %d previews on system path '%s'",
                 size, systemAppDataPath.c_str());
         Log << message;
-            
+
         Screen *scr = video::GetScreen();
         GC gc (scr->get_surface());
         Font *f = enigma::GetFont("menufont");
@@ -390,35 +491,48 @@ void Application::init(int argc, char **argv)
         vline(gc, 470, 280, 20);
         scr->update_all ();
         scr->flush_updates();
-        
+
         int i = 0;
-        for (it = proxies.begin(); it != proxies.end(); it++, i++) {
-            gui::LevelPreviewCache::makeSystemPreview(*it, systemAppDataPath);
-            vline(gc, 170 + i*300 / size, 280, 20);
-            scr->update_all ();
-            scr->flush_updates();
+        for (int m=0; m<3; m++) {
+            switch (m) {
+                case 0 : video::SetThumbInfo(120, 78, "-120x78"); break;
+                case 1 : video::SetThumbInfo(160, 104, "-160x104"); break;
+                case 2 : video::SetThumbInfo(60, 39, "-60x39"); break;
+            }
+            for (it = proxies.begin(); it != proxies.end(); it++, i++) {
+                Log << "Make preview " << (*it)->getId() << "\n";
+                gui::LevelPreviewCache::makeSystemPreview(*it, systemAppDataPath);
+                // i counts from 0 to 3*size (3 video modes),
+                // this makes up for the factor 100 = 300 / 3.
+                vline(gc, 170 + i*100 / size, 280, 20);
+                scr->update_all ();
+                scr->flush_updates();
+            }
         }
+        Log << "Make preview finished succesfully\n";
         return;
     }
-    
+
     // initialize score -- needs random init
     lev::ScoreManager::instance();
 }
 
 std::string Application::getVersionInfo() {
     std::string versionInfo;
-    double enigmaVersion;
-    sscanf(PACKAGE_VERSION,"%4lf",&enigmaVersion);
     if (enigmaVersion >= ENIGMACOMPATIBITLITY)
         versionInfo = "v" PACKAGE_VERSION;
     else {
-        versionInfo =  "v" PACKAGE_VERSION 
-            " (development version - v" + 
-            ecl::strf("%.2f",ENIGMACOMPATIBITLITY) + " compatibilty branch)";
+        versionInfo =  "v" PACKAGE_VERSION
+            " (development version - v" +
+            ecl::strf("%.2f",ENIGMACOMPATIBITLITY) + " compatibility branch)";
     }
     return versionInfo;
 }
 
+double Application::getEnigmaVersion() {
+    return enigmaVersion;
+}
+
 void Application::initSysDatapaths(const std::string &prefFilename)
 {
     std::string progDir;          // directory path part of args[0]
@@ -426,13 +540,13 @@ void Application::initSysDatapaths(const std::string &prefFilename)
     bool progDirExists = split_path(progCallPath, &progDir, &progName);
     std::string systemPath = SYSTEM_DATA_DIR; // substituted by configure.ac
     bool haveHome = (getenv("HOME") != 0);
+
+
 #ifdef __MINGW32__
     // windows standard user specific application data directory path
     std::string winAppDataPath = ecl::ApplicationDataPath() + "/Enigma";
 #endif
 
-    // systemFS
-    systemFS = new GameFS();
 #ifdef __MINGW32__
     if (!progDirExists) {
         // filename only -- working dir should be on enigma as enigma
@@ -452,18 +566,30 @@ void Application::initSysDatapaths(const std::string &prefFilename)
     // then chdir to ../Resources. The original SDL implementation chdirs to
     // "../../..", i.e. the directory the bundle is placed in. This breaks
     // the self-containedness.
-	
+
     systemAppDataPath = progDir + "/../Resources/data";
-    
+
 #else
     // Unix -- we get our data path from the installation
     systemAppDataPath = systemPath;
 #endif
+
+    // systemFS
+    systemFS = new GameFS();
     systemFS->append_dir(systemAppDataPath);
     if (!systemCmdDataPath.empty())
          systemFS->prepend_dir(systemCmdDataPath);
-    Log << "systemFS = \"" << systemFS->getDataPath() << "\"\n"; 
-    
+
+    // docPath
+    docPath = DOCDIR;    // defined in src/Makefile.am
+#ifdef __MINGW32__
+    if (progDirExists) {
+        docPath = progDir;
+    }
+#elif MACOSX
+    docPath = progDir + "/../Resources/doc";
+#endif
+
     // l10nPath
     l10nPath = LOCALEDIR;    // defined in src/Makefile.am
 #ifdef __MINGW32__
@@ -473,16 +599,14 @@ void Application::initSysDatapaths(const std::string &prefFilename)
 #elif MACOSX
     l10nPath = progDir + "/../Resources/locale";
 #endif
-    Log << "l10nPath = \"" << l10nPath << "\"\n"; 
-    
-    
+
     // prefPath
     if (prefFilename.find_first_of(ecl::PathSeparators) != std::string::npos) {
         // pref is a path - absolute or home relative
         prefPath = ecl::ExpandPath(prefFilename);
         if (!ecl::FolderExists(prefPath))
             if(!ecl::FolderCreate(prefPath)) {
-                fprintf(stderr, ("Error cannot create pref directory.\n"));
+                fprintf(stderr, ("Error: Cannot create pref directory.\n"));
                 exit(1);
             }
         userStdPath = prefPath; // default if pref is a path
@@ -492,7 +616,7 @@ void Application::initSysDatapaths(const std::string &prefFilename)
         if (!ecl::FolderExists(prefPath))
             // may happen on Windows
             if(!ecl::FolderCreate(prefPath)) {
-                fprintf(stderr, _("Error Home directory does not exist.\n"));
+                fprintf(stderr, _("Error: Home directory does not exist.\n"));
                 exit(1);
             }
 #ifdef MACOSX
@@ -507,18 +631,17 @@ void Application::initSysDatapaths(const std::string &prefFilename)
         if (!ecl::FolderExists(winAppDataPath))
             // may happen on Windows
             if(!ecl::FolderCreate(winAppDataPath)) {
-                fprintf(stderr, _("Error Application Data directory does not exist.\n"));
+                fprintf(stderr, _("Error: Application Data directory does not exist.\n"));
                 exit(1);
             }
-        Log << "winAppDataPath " << winAppDataPath << "\n";
+//        Log << "winAppDataPath " << winAppDataPath << "\n";
         userStdPath = winAppDataPath;
         prefPath = winAppDataPath + ecl::PathSeparator + "." + prefFilename;
 #endif
     } else {
-        fprintf(stderr, _("Error Home directory does not exist.\n"));
+        fprintf(stderr, _("Error: Home directory does not exist.\n"));
         exit(1);
     }
-    Log << "prefPath = \"" << prefPath << "\"\n"; 
 }
 
 void Application::initXerces() {
@@ -528,17 +651,17 @@ void Application::initXerces() {
         // If more than the error messages should be influenced we would
         // have to terminate and reinit after reading the user prefs.
         XMLPlatformUtils::Initialize();
-        
+
         // Initialize transcoding service for XML <-> utf8
-        XMLTransService::Codes initResult;        
+        XMLTransService::Codes initResult;
         xmlUtf8Transcoder = XMLPlatformUtils::fgTransService->
-                makeNewTranscoderFor(XMLRecognizer::UTF_8, initResult, 
-                4096); // the block size is irrelevant for utf-8                  
+                makeNewTranscoderFor(XMLRecognizer::UTF_8, initResult,
+                4096); // the block size is irrelevant for utf-8
         if (initResult != XMLTransService::Ok) {
             fprintf(stderr, _("Error in XML initialization.\n"));
             exit(1);
         }
-        
+
         static const XMLCh ls[] = { chLatin_L, chLatin_S, chNull };
         static const XMLCh core[] = { chLatin_C, chLatin_O,  chLatin_R, chLatin_E, chNull };
         domImplementationLS = (DOMImplementationLS*)
@@ -563,11 +686,11 @@ void Application::initXerces() {
 
         domSer = domImplementationLS->createLSSerializer();
         config = domSer->getDomConfig();
-        
+
         config->setParameter(XMLUni::fgDOMWRTFormatPrettyPrint, true);
         config->setParameter(XMLUni::fgDOMErrorHandler, domSerErrorHandler);
 
-#else    
+#else
         domParser = domImplementationLS->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0);
 
         domParser->setFeature(XMLUni::fgDOMNamespaces, true);
@@ -586,8 +709,8 @@ void Application::initXerces() {
     catch (...) {
         fprintf(stderr, _("Error in XML initialization.\n"));
         exit(1);
-    }     
-}	
+    }
+}
 
 void Application::initUserDatapaths() {
     // userPath
@@ -609,24 +732,24 @@ void Application::initUserDatapaths() {
     } else {
         userPath = XMLtoLocal(Utf8ToXML(userPath.c_str()).x_str()).c_str();
     }
-    Log << "userPath = \"" << userPath << "\"\n"; 
-    
+    Log << "userPath = \"" << userPath << "\"\n";
+
     // userImagePath
     userImagePath = prefs->getString("UserImagePath");
     if (userImagePath.empty()) {
 #ifdef MACOSX
-        userImagePath = userStdPathMac1_00;  // empty prefs user path is 1.00 std user path 
+        userImagePath = userStdPathMac1_00;  // empty prefs user path is 1.00 std user path
 #else
         userImagePath = userStdPath;
 #endif
     } else {
         userImagePath = XMLtoLocal(Utf8ToXML(userImagePath.c_str()).x_str()).c_str();
     }
-    Log << "userImagePath = \"" << userImagePath << "\"\n"; 
+    Log << "userImagePath = \"" << userImagePath << "\"\n";
 
     // resourceFS
     resourceFS = new GameFS();
-    resourceFS->append_dir(systemAppDataPath);    
+    resourceFS->append_dir(systemAppDataPath);
 #ifdef MACOSX
     // set user-visible data paths -- use it for resource paths
     NSSearchPathEnumerationState cur=NSStartSearchPathEnumeration(NSLibraryDirectory, NSAllDomainsMask);
@@ -638,26 +761,28 @@ void Application::initUserDatapaths() {
     }
 #endif
     if (!systemCmdDataPath.empty())
-	resourceFS->prepend_dir(systemCmdDataPath);    
+	resourceFS->prepend_dir(systemCmdDataPath);
     resourceFS->prepend_dir(userPath);
     if (userImagePath != userPath)
         resourceFS->prepend_dir(userImagePath);
-    Log << "resourceFS = \"" << resourceFS->getDataPath() << "\"\n"; 
+    Log << "resourceFS = \"" << resourceFS->getDataPath() << "\"\n";
 
     // create levels/auto, levels/cross, levels/legacy_dat on userPath
     if (!ecl::FolderExists(userPath + "/levels/auto"))
         ecl::FolderCreate (userPath + "/levels/auto");
     if (!ecl::FolderExists(userPath + "/levels/cross"))
         ecl::FolderCreate (userPath + "/levels/cross");
+    if (!ecl::FolderExists(userPath + "/levels/externaldata"))
+        ecl::FolderCreate (userPath + "/levels/externaldata");
     if (!ecl::FolderExists(userPath + "/levels/legacy_dat"))
-        ecl::FolderCreate (userPath + "/levels/legacy_dat");   
+        ecl::FolderCreate (userPath + "/levels/legacy_dat");
     if (!ecl::FolderExists(userPath + "/backup"))
-        ecl::FolderCreate (userPath + "/backup");   
+        ecl::FolderCreate (userPath + "/backup");
 }
 
 #ifdef MACOSX
 void Application::updateMac1_00() {
-    if (prefs->getInt("_MacUpdate1.00") == 0 && 
+    if (prefs->getInt("_MacUpdate1.00") == 0 &&
             prefs->getString("UserPath").empty() &&
             prefs->getString("UserImagePath").empty()) {
         gui::ErrorMenu m(ecl::strf(N_("Mac OS X upgrade from Enigma 1.00\n\nThe default user data path has changed from\n  %s \n\nto the visible data path\n  %s \n\nIf ok Enigma will move your data to this new location.\nNote that you have to restart Enigma once for completion of this update."), userStdPathMac1_00.c_str(), userStdPath.c_str()),
@@ -669,7 +794,7 @@ void Application::updateMac1_00() {
             prefs->setProperty("_MacUpdate1.00", 0);
         } else {  // OK move now
             Log << "Mac update\n";
-            // move 
+            // move
             std::system(ecl::strf("mkdir '%s' && cd ~/.enigma && tar -cf - * | (cd '%s' && tar -xf -) && cd ~ && rm -rf ~/.enigma", userStdPath.c_str(), userStdPath.c_str()).c_str());
             setUserPath("");
             setUserImagePath("");
@@ -684,7 +809,7 @@ void Application::updateMac1_00() {
 void Application::init_i18n()
 {
     // Initialize the internationalization subsystem
-    
+
     // priorities:
     // language: command-line --- user option --- system (environment)
     // defaultLanguage: command-line --- system (environment)
@@ -698,7 +823,7 @@ void Application::init_i18n()
         if (app.language == "") {
             app.language = app.defaultLanguage;
         }
-    }   
+    }
 
 #if defined(ENABLE_NLS)
 
@@ -715,12 +840,7 @@ void Application::init_i18n()
 
 void Application::setLanguage(std::string newLanguage)
 {
-    if (newLanguage == "") {
-        language = defaultLanguage;
-    }
-    else {
-        language = newLanguage;
-    }
+    language = newLanguage.empty() ? defaultLanguage : newLanguage;
     nls::SetMessageLocale(language);
 }
 
@@ -728,15 +848,15 @@ void Application::setUserPath(std::string newPath) {
     std::string prefUserPath = (newPath == userStdPath) ? "" : newPath;
     if ((prefUserPath.empty() && userPath != userStdPath) || (!prefUserPath.empty() && prefUserPath != userPath)) {
         // set the new userPath - used for saves
-        if (prefUserPath.empty())    
+        if (prefUserPath.empty())
             userPath = userStdPath;
         else
             userPath = prefUserPath;
-        
+
         // load all resources primarily from the new path but keep the old user path
         // because the user could not yet copy his user data to the new location
         resourceFS->prepend_dir(userPath);
-        
+
         // set the new path as the users preference - the standard path is saved as ""
 #ifdef MACOSX
         // 1.00 uses "" as "~/.enigma" - we have to store the complete path
@@ -750,11 +870,11 @@ void Application::setUserImagePath(std::string newPath) {
     std::string prefUserImagePath = (newPath == userStdPath) ? "" : newPath;
     if ((prefUserImagePath.empty() && userImagePath != userStdPath) || (!prefUserImagePath.empty() && prefUserImagePath != userImagePath)) {
         // set the new userImagePath - used for saves
-        if (prefUserImagePath.empty())    
+        if (prefUserImagePath.empty())
             userImagePath = userStdPath;
         else
             userImagePath = prefUserImagePath;
-                
+
         // load all resources primarily from the new path but keep the old user path
         // because the user could not yet copy his user data to the new location
         if (userImagePath != userPath)
@@ -771,14 +891,14 @@ void Application::setUserImagePath(std::string newPath) {
 
 /* -------------------- Functions -------------------- */
 
-void Application::shutdown() 
+void Application::shutdown()
 {
     oxyd::Shutdown();
-    world::Shutdown();
+    enigma::ShutdownWorld();
     display::Shutdown();
     if (!isMakePreviews) { // avoid saves on preview generation
         lev::RatingManager::instance()->save();
-        if (lev::PersistentIndex::historyIndex != NULL) 
+        if (lev::PersistentIndex::historyIndex != NULL)
             lev::PersistentIndex::historyIndex->save();
         lev::ScoreManager::instance()->shutdown();
         app.state->shutdown();
@@ -789,12 +909,19 @@ void Application::shutdown()
     video::Shutdown();
     sound::Shutdown();
     enet_deinitialize();
+    enigma::ShutdownCurl();
     lua::ShutdownGlobal();
     XMLPlatformUtils::Terminate();
+#ifdef SDL_IMG_INIT
+    IMG_Quit();
+#endif
+    ClearFontCache();
+    TTF_Quit();
+    ClearImageCache();
     delete ::nullbuffer;
 }
 
-int main(int argc, char** argv) 
+int main(int argc, char **argv)
 {
     try {
         app.init(argc,argv);
@@ -805,7 +932,7 @@ int main(int argc, char** argv)
     }
     catch (XFrontend &e) {
         cerr << "Error: " << e.what() << endl;
-        std::string message = _("Fatal Error that causes the application to quit:\n\n");
+        std::string message = _("Fatal error that caused the application to quit:\n\n");
         if (app.errorInit) {
             gui::ErrorMenu m(message + e.what(), N_("Quit"));
             m.manage();
